#include "pch.h"
#include "BasicFingerPaintRenderer.h"
#include "Common/DirectXHelper.h"	// For ThrowIfFailed

using namespace BasicFingerPaint;
using namespace D2D1;
using namespace Microsoft::WRL;
using namespace Platform;       // for COMException
using namespace Windows::Foundation::Collections;

BasicFingerPaintRenderer::BasicFingerPaintRenderer(const std::shared_ptr<DeviceResources>& deviceResources) :
    m_deviceResources(deviceResources)
{
    // Create resources
    CreateDeviceDependentResources();
}

void BasicFingerPaintRenderer::CreateDeviceDependentResources()
{
    // This is actually a device-independent resource
    D2D1_STROKE_STYLE_PROPERTIES strokeStyleProps;
    strokeStyleProps.dashStyle = D2D1_DASH_STYLE_SOLID;
    strokeStyleProps.dashCap = D2D1_CAP_STYLE_ROUND;
    strokeStyleProps.miterLimit = 10;
    strokeStyleProps.dashOffset = 0;
    strokeStyleProps.startCap = D2D1_CAP_STYLE_ROUND;
    strokeStyleProps.endCap = D2D1_CAP_STYLE_ROUND;
    strokeStyleProps.lineJoin = D2D1_LINE_JOIN_ROUND;

    DX::ThrowIfFailed(m_deviceResources->GetD2DFactory()->
        CreateStrokeStyle(strokeStyleProps, nullptr, 0, &m_strokeStyle)
        );

    // So is this
    DX::ThrowIfFailed(
        m_deviceResources->GetD2DFactory()->CreateDrawingStateBlock(&m_stateBlock)
        );

    DX::ThrowIfFailed(
        m_deviceResources->GetD2DDeviceContext()->CreateSolidColorBrush(D2D1::ColorF(D2D1::ColorF::Black), &m_solidColorBrush)
        );

    for (StrokeInfo& strokeInfo : completedStrokes)
        strokeInfo.GeometryRealization = CreateGeometryRealization(strokeInfo);
}

void BasicFingerPaintRenderer::ReleaseDeviceDependentResources()
{
    m_strokeStyle.Reset();
    m_stateBlock.Reset();
    m_solidColorBrush.Reset();

    for (StrokeInfo& strokeInfo : completedStrokes)
        strokeInfo.GeometryRealization.Reset();
}

void BasicFingerPaintRenderer::SetStrokes(std::vector<StrokeInfo> strokes)
{
    completedStrokes = strokes;

    for (StrokeInfo& strokeInfo : completedStrokes)
    {
        strokeInfo.Geometry = CreatePolylinePathGeometry(strokeInfo.Points);
        strokeInfo.GeometryRealization = CreateGeometryRealization(strokeInfo);
    }
}

std::vector<StrokeInfo> BasicFingerPaintRenderer::GetStrokes()
{
    return completedStrokes;
}

void BasicFingerPaintRenderer::BeginStroke(unsigned int id, D2D1_POINT_2F point, float width, ColorF color)
{
    // Save stroke information in StrokeInfo structure
    StrokeInfo strokeInfo;
    strokeInfo.Points.push_back(point);
    strokeInfo.Color = color;
    strokeInfo.Width = width;

    // Store in map with ID number
    strokesInProgress.insert(std::make_pair(id, strokeInfo));
}

void BasicFingerPaintRenderer::ContinueStroke(unsigned int id, D2D1_POINT_2F point)
{
    // Get the iterator for strokesInProgress
    auto const it = strokesInProgress.find(id);

    // Never started a stroke, so skip
    if (it == strokesInProgress.end())
        return;

    // Get the StrokeInfo object for this finger
    StrokeInfo& strokeInfo = it->second;
    D2D1_POINT_2F previousPoint = strokeInfo.Points.back();

    // Skip duplicate points
    if (point.x != previousPoint.x || point.y != previousPoint.y)
    {
        strokeInfo.Points.push_back(point);
        strokeInfo.Geometry = CreatePolylinePathGeometry(strokeInfo.Points);
    }
}

void BasicFingerPaintRenderer::EndStroke(unsigned int id, D2D1_POINT_2F point)
{
    auto const it = strokesInProgress.find(id);

    // Never started a stroke, so skip
    if (it == strokesInProgress.end())
        return;

    // Get the StrokeInfo object for this finger
    StrokeInfo strokeInfo = it->second;

    // Save new point
    strokeInfo.Points.push_back(point);
    strokeInfo.Geometry = CreatePolylinePathGeometry(strokeInfo.Points);

    // Create realized geometry
    strokeInfo.GeometryRealization = CreateGeometryRealization(strokeInfo);

    // Remove from map, save in vector
    strokesInProgress.erase(it);
    completedStrokes.emplace_back(std::move(strokeInfo));
}

void BasicFingerPaintRenderer::CancelStroke(unsigned int id)
{
    if (strokesInProgress.count(id) > 0)
    {
        strokesInProgress.erase(id);
    }
}

void BasicFingerPaintRenderer::ClearAll()
{
    completedStrokes.clear();
}

// Renders a frame to the screen.
void BasicFingerPaintRenderer::Render()
{
    ID2D1DeviceContext1* pContext = m_deviceResources->GetD2DDeviceContext();

    pContext->SaveDrawingState(m_stateBlock.Get());
    pContext->BeginDraw();
    pContext->Clear(ColorF(ColorF::White));
    pContext->SetTransform(m_deviceResources->GetOrientationTransform2D());

    // First render the completed strokes
    for (const StrokeInfo& strokeInfo : completedStrokes)
    {
        m_solidColorBrush->SetColor(strokeInfo.Color);

        pContext->DrawGeometryRealization(strokeInfo.GeometryRealization.Get(),
                                          m_solidColorBrush.Get());
    }

    // Then render the strokes in progress
    for (const std::pair<unsigned int, StrokeInfo>& pair : strokesInProgress)
    {
        StrokeInfo strokeInfo = pair.second;

        if (strokeInfo.Geometry != nullptr)
        {
            m_solidColorBrush->SetColor(strokeInfo.Color);

            pContext->DrawGeometry(strokeInfo.Geometry.Get(),
                                   m_solidColorBrush.Get(),
                                   strokeInfo.Width,
                                   m_strokeStyle.Get());
        }
    }

    // We ignore D2DERR_RECREATE_TARGET here. This error indicates that the device
    // is lost. It will be handled during the next call to Present.
    HRESULT hr = pContext->EndDraw();
    if (hr != D2DERR_RECREATE_TARGET)
    {
        DX::ThrowIfFailed(hr);
    }

    pContext->RestoreDrawingState(m_stateBlock.Get());
}

ComPtr<ID2D1PathGeometry> BasicFingerPaintRenderer::CreatePolylinePathGeometry(const std::vector<D2D1_POINT_2F>& points)
{
    // Create the PathGeometry
    ComPtr<ID2D1PathGeometry> pathGeometry;
    HRESULT hresult = m_deviceResources->GetD2DFactory()->CreatePathGeometry(&pathGeometry);

    if (FAILED(hresult))
        throw ref new COMException(hresult, "factory->CreatePathGeometry failure");

    // Get the GeometrySink of the PathGeometry
    ComPtr<ID2D1GeometrySink> geometrySink;
    hresult = pathGeometry->Open(&geometrySink);

    if (FAILED(hresult))
        throw ref new COMException(hresult, "pathGeometry->Open failure");

    // Begin figure, add lines, end figure, and close
    geometrySink->BeginFigure(points.at(0), D2D1_FIGURE_BEGIN_HOLLOW);
    geometrySink->AddLines(points.data() + 1, points.size() - 1);
    geometrySink->EndFigure(D2D1_FIGURE_END_OPEN);
    hresult = geometrySink->Close();

    if (FAILED(hresult))
        throw ref new COMException(hresult, "geometrySink->Close failure");

    return pathGeometry;
}

ComPtr<ID2D1GeometryRealization> BasicFingerPaintRenderer::CreateGeometryRealization(StrokeInfo& strokeInfo)
{
    ComPtr<ID2D1GeometryRealization> geometryRealization;

    HRESULT hr = m_deviceResources->GetD2DDeviceContext()->CreateStrokedGeometryRealization(strokeInfo.Geometry.Get(),
                                                                                            D2D1_DEFAULT_FLATTENING_TOLERANCE,
                                                                                            strokeInfo.Width,
                                                                                            m_strokeStyle.Get(),
                                                                                            &geometryRealization);

    if (FAILED(hr))
        throw ref new COMException(hr, "deviceContext->CreateStrokedGeometryRealization failure");
    
    return geometryRealization;
}
