﻿//
// MainPage.xaml.cpp
// Implementation of the MainPage class.
//

#include "pch.h"
#include "DirectXPage.xaml.h"

using namespace BasicFingerPaint;

using namespace Concurrency;
using namespace Platform;
using namespace Platform::Collections;
using namespace Windows::ApplicationModel;
using namespace Windows::Foundation;
using namespace Windows::Foundation::Collections;
using namespace Windows::Graphics::Display;
using namespace Windows::Storage;
using namespace Windows::Storage::Search;
using namespace Windows::Storage::Streams;
using namespace Windows::System;
using namespace Windows::System::Threading;
using namespace Windows::UI;
using namespace Windows::UI::Core;
using namespace Windows::UI::Input;
using namespace Windows::UI::Popups;
using namespace Windows::UI::Xaml;
using namespace Windows::UI::Xaml::Controls;
using namespace Windows::UI::Xaml::Controls::Primitives;
using namespace Windows::UI::Xaml::Data;
using namespace Windows::UI::Xaml::Input;
using namespace Windows::UI::Xaml::Media;
using namespace Windows::UI::Xaml::Navigation;

String^ DirectXPage::currentDrawingFilename = "CurrentDrawing";
String^ DirectXPage::filenameExtension = ".drawing";

DirectXPage::DirectXPage():
	m_windowVisible(true)
{
	InitializeComponent();

    // Default ComboBox selections
    colorComboBox->SelectedItem = NamedColor::FromName("Black");
    widthComboBox->SelectedIndex = 5;

    // Possibly enable Open button
    task<IVectorView<String^>^> getFilesTask = create_task(GetFileNamesAsync());
    getFilesTask.then([this](IVectorView<String^>^ filenames)
    {
        appBarOpenButton->IsEnabled = filenames->Size > 0;
    });

	// Register event handlers for page lifecycle.
	CoreWindow^ window = Window::Current->CoreWindow;

	window->VisibilityChanged +=
		ref new TypedEventHandler<CoreWindow^, VisibilityChangedEventArgs^>(this, &DirectXPage::OnVisibilityChanged);

	DisplayInformation^ currentDisplayInformation = DisplayInformation::GetForCurrentView();

	currentDisplayInformation->DpiChanged +=
		ref new TypedEventHandler<DisplayInformation^, Object^>(this, &DirectXPage::OnDpiChanged);

	currentDisplayInformation->OrientationChanged +=
		ref new TypedEventHandler<DisplayInformation^, Object^>(this, &DirectXPage::OnOrientationChanged);

	DisplayInformation::DisplayContentsInvalidated +=
		ref new TypedEventHandler<DisplayInformation^, Object^>(this, &DirectXPage::OnDisplayContentsInvalidated);

	swapChainPanel->CompositionScaleChanged += 
		ref new TypedEventHandler<SwapChainPanel^, Object^>(this, &DirectXPage::OnCompositionScaleChanged);

	swapChainPanel->SizeChanged +=
		ref new SizeChangedEventHandler(this, &DirectXPage::OnSwapChainPanelSizeChanged);

    // --------------------------------------------
    // CompositionTarget::Rendering event not used!
    // --------------------------------------------
	// Register the rendering event, called every time XAML renders the screen.
    // m_eventToken = CompositionTarget::Rendering::add(ref new EventHandler<Object^>(this, &DirectXPage::OnRendering));

	// Whenever this screen is not being used anymore, you can unregister this event with the following line:
	// CompositionTarget::Rendering::remove(m_eventToken);

	// Disable all pointer visual feedback for better performance when touching.
	auto pointerVisualizationSettings = PointerVisualizationSettings::GetForCurrentView();
	pointerVisualizationSettings->IsContactFeedbackEnabled = false; 
	pointerVisualizationSettings->IsBarrelButtonFeedbackEnabled = false;

	// At this point we have access to the device. 
	// We can create the device-dependent resources.
	m_deviceResources = std::make_shared<DeviceResources>();
	m_deviceResources->SetWindow(Window::Current->CoreWindow, swapChainPanel);

	// Register our SwapChainPanel to get independent input pointer events
	auto workItemHandler = ref new WorkItemHandler([this] (IAsyncAction ^)
	{
		// The CoreIndependentInputSource will raise pointer events for the specified device types on whichever thread it's created on.
		CoreIndependentInputSource^ coreInput = swapChainPanel->CreateCoreIndependentInputSource(
			Windows::UI::Core::CoreInputDeviceTypes::Mouse |
			Windows::UI::Core::CoreInputDeviceTypes::Touch |
			Windows::UI::Core::CoreInputDeviceTypes::Pen
			);

		// Register for pointer events, which will be raised on the background thread.
		coreInput->PointerPressed += ref new TypedEventHandler<Object^, PointerEventArgs^>(this, &DirectXPage::OnPointerPressed);
        coreInput->PointerMoved += ref new TypedEventHandler<Object^, PointerEventArgs^>(this, &DirectXPage::OnPointerMoved);
        coreInput->PointerReleased += ref new TypedEventHandler<Object^, PointerEventArgs^>(this, &DirectXPage::OnPointerReleased);

		// Begin processing input messages as they're delivered.
        coreInput->Dispatcher->ProcessEvents(CoreProcessEventsOption::ProcessUntilQuit);
	});

	// Run task on a dedicated high priority background thread.
	m_inputLoopWorker = ThreadPool::RunAsync(workItemHandler, WorkItemPriority::High, WorkItemOptions::TimeSliced);


	m_main = std::unique_ptr<BasicFingerPaintMain>(new BasicFingerPaintMain(m_deviceResources));
}

// Saves the current state of the app for suspend and terminate events.
void DirectXPage::SaveInternalState(SuspendingOperation^ suspendingOperation, IPropertySet^ state)
{
    critical_section::scoped_lock lock(m_criticalSection);
    m_deviceResources->Trim();

	// Save program settings
    state->Clear();

    NamedColor^ namedColor = dynamic_cast<NamedColor^>(colorComboBox->SelectedItem);
    if (namedColor != nullptr)
        state->Insert("Color", namedColor->Name);

    if (widthComboBox->SelectedIndex != -1)
        state->Insert("Width", (int) (widthComboBox->SelectedItem));

    // Save current drawing
    IVector<StrokeInformation^>^ strokes = m_main->GetStrokes();
    SuspendingDeferral^ deferral = suspendingOperation->GetDeferral();

    task<void> saveFileTask = create_task(SaveDrawingAsync(strokes, DirectXPage::currentDrawingFilename));
    saveFileTask.then([deferral]()
    {
        deferral->Complete();
    });
}

// Loads the current state of the app for resume events.
void DirectXPage::LoadInternalState(IPropertySet^ state)
{
	// Load saved settings
    if (state->HasKey("Color"))
    {
        String^ colorName = safe_cast<String^> (state->Lookup("Color"));
        colorComboBox->SelectedItem = NamedColor::FromName(colorName);
    }

    if (state->HasKey("Width"))
    {
        widthComboBox->SelectedItem = (int) (state->Lookup("Width"));
    }

    // Load saved drawing
    task<IVector<StrokeInformation^>^> openDrawingTask = create_task(OpenDrawingAsync(DirectXPage::currentDrawingFilename));
    openDrawingTask.then([this] (IVector<StrokeInformation^>^ strokes)
    {
        critical_section::scoped_lock lock(m_criticalSection);
        m_main->SetStrokes(strokes);
        m_main->Render();
        m_deviceResources->Present();
    });
}

// ---------
// Disabled!
// ---------
// Called every time XAML decides to render a frame.
void DirectXPage::OnRendering(Object^ sender, Object^ args)
{
	if (m_windowVisible)
	{
		m_main->Update();
        
		if (m_main->Render())
		{
			m_deviceResources->Present();
		}
	}
}

// Window event handlers.

void DirectXPage::OnVisibilityChanged(CoreWindow^ sender, VisibilityChangedEventArgs^ args)
{
	m_windowVisible = args->Visible;
}

// Display properties event handlers.

void DirectXPage::OnDpiChanged(DisplayInformation^ sender, Object^ args)
{
    critical_section::scoped_lock lock(m_criticalSection);
    m_deviceResources->SetDpi(sender->LogicalDpi);
}


void DirectXPage::OnOrientationChanged(DisplayInformation^ sender, Object^ args)
{
    critical_section::scoped_lock lock(m_criticalSection);
    m_deviceResources->UpdateForWindowSizeChange();
	m_main->CreateWindowSizeDependentResources();
}


void DirectXPage::OnDisplayContentsInvalidated(DisplayInformation^ sender, Object^ args)
{
    critical_section::scoped_lock lock(m_criticalSection);
    m_deviceResources->ValidateDevice();
}

void DirectXPage::OnPointerPressed(Object^ sender, PointerEventArgs^ args)
{
    critical_section::scoped_lock lock(m_criticalSection);

    m_main->BeginStroke(args->CurrentPoint->PointerId, 
                        args->CurrentPoint->Position, 
                        m_selectedWidth, 
                        m_selectedColor);

    m_main->Render();
    m_deviceResources->Present();
}

void DirectXPage::OnPointerMoved(Object^ sender, PointerEventArgs^ args)
{
    critical_section::scoped_lock lock(m_criticalSection);

    IVector<PointerPoint^>^ points = args->GetIntermediatePoints();

    for (auto pPoint = end(points) - 1; pPoint >= begin(points); pPoint--)
    {
        m_main->ContinueStroke((*pPoint)->PointerId, (*pPoint)->Position);
    }

    m_main->Render();
    m_deviceResources->Present();
}

void DirectXPage::OnPointerReleased(Object^ sender, PointerEventArgs^ args)
{
    critical_section::scoped_lock lock(m_criticalSection);

    m_main->EndStroke(args->CurrentPoint->PointerId, args->CurrentPoint->Position);

    m_main->Render();
    m_deviceResources->Present();
}

void DirectXPage::OnCompositionScaleChanged(SwapChainPanel^ sender, Object^ args)
{
    critical_section::scoped_lock lock(m_criticalSection);

    m_deviceResources->CreateWindowSizeDependentResources();
	m_main->CreateWindowSizeDependentResources();

    m_main->Render();
    m_deviceResources->Present();
}

void DirectXPage::OnSwapChainPanelSizeChanged(Object^ sender, SizeChangedEventArgs^ e)
{
    critical_section::scoped_lock lock(m_criticalSection);

    m_deviceResources->UpdateForWindowSizeChange();
	m_main->CreateWindowSizeDependentResources();

    m_main->Render();
    m_deviceResources->Present();
}


void DirectXPage::OnColorComboBoxSelectionChanged(Object^ sender, SelectionChangedEventArgs^ args)
{
    NamedColor^ namedColor = dynamic_cast<NamedColor^>(colorComboBox->SelectedItem);
    m_selectedColor = namedColor != nullptr ? namedColor->Color : Colors::Black;
}

void DirectXPage::OnWidthComboBoxSelectionChanged(Object^ sender, SelectionChangedEventArgs^ args)
{
    m_selectedWidth = (float)(widthComboBox->SelectedIndex != -1 ? (int) widthComboBox->SelectedItem : 5);
}

void DirectXPage::OnClearAllAppBarButtonClick(Object^ sender, RoutedEventArgs^ args)
{
    critical_section::scoped_lock lock(m_criticalSection);

    m_main->ClearAll();
    m_main->Render();
    m_deviceResources->Present();
}

void DirectXPage::OnOpenFileFlyoutOpening(Object^ sender, Object^ args)
{
    task<IVectorView<String^>^> getFilesTask = create_task(GetFileNamesAsync());
    getFilesTask.then([this](IVectorView<String^>^ filenames)
    {
        openFileListBox->ItemsSource = filenames;
    });
}

void DirectXPage::OnOpenFileListBoxSelectionChanged(Object^ sender, SelectionChangedEventArgs^ args)
{
    openFileOpenButton->IsEnabled = safe_cast<ListBox^>(sender)->SelectedItem != nullptr;
}

void DirectXPage::OnOpenFileOpenButtonClick(Object^ sender, RoutedEventArgs^ args)
{
    openFileFlyout->Hide();
    String^ filename = safe_cast<String^>(openFileListBox->SelectedItem);

    task<IVector<StrokeInformation^>^> openDrawingTask = create_task(OpenDrawingAsync(filename + DirectXPage::filenameExtension));
    openDrawingTask.then([this](IVector<StrokeInformation^>^ strokes)
    {
        critical_section::scoped_lock lock(m_criticalSection);
        m_main->SetStrokes(strokes);
        m_main->Render();
        m_deviceResources->Present();
    });
}

task<IVector<StrokeInformation^>^> DirectXPage::OpenDrawingAsync(String^ filename)
{
    StorageFolder^ folder = ApplicationData::Current->LocalFolder;
    IVector<StrokeInformation^>^ strokes = ref new Vector<StrokeInformation^>();

    task<StorageFile^> createFileTask = create_task(folder->CreateFileAsync(filename, CreationCollisionOption::OpenIfExists));
    return createFileTask.then([](StorageFile^ file)
    {
        return file->OpenAsync(FileAccessMode::ReadWrite);
    })
        .then([this, strokes](IRandomAccessStream^ stream)
    {
        if (stream->Size == 0)
            return strokes;

        DataReader^ reader = ref new DataReader(stream);
        task<unsigned int> loadTask = create_task(reader->LoadAsync(unsigned int(stream->Size)));
        loadTask.then([this, reader, strokes](unsigned int)
        {
            unsigned int numStrokes = reader->ReadUInt32();

            for (unsigned int i = 0; i < numStrokes; i++)
            {
                StrokeInformation^ strokeInformation = ref new StrokeInformation();
                strokeInformation->Width = reader->ReadSingle();
                byte a = reader->ReadByte();
                byte r = reader->ReadByte();
                byte g = reader->ReadByte();
                byte b = reader->ReadByte();
                strokeInformation->Color = ColorHelper::FromArgb(a, r, g, b);

                unsigned int numPoints = reader->ReadUInt32();

                for (unsigned int i = 0; i < numPoints; i++)
                {
                    float x = reader->ReadSingle();
                    float y = reader->ReadSingle();
                    strokeInformation->Points->Append(Point(x, y));
                }
                strokes->Append(strokeInformation);
            }
            return strokes;
        });

        return strokes;
    });
}

void DirectXPage::OnSaveAsFlyoutOpening(Object^ sender, Object^ args)
{
    // Come up with unique filename
    task<IVectorView<String^>^> getFilesTask = create_task(GetFileNamesAsync());
    getFilesTask.then([this](IVectorView<String^>^ fileNames)
    {
        int number = 0;
        bool isUnique;

        do
        {
            number += 1;
            isUnique = true;
            String^ name = "drawing " + number;

            for (String^ fileName : fileNames)
            {
                if (name == fileName)
                {
                    isUnique = false;
                    break;
                }
            }
        } while (!isUnique);

        // Create Save dialog
        saveTextBox->Text = "drawing " + number.ToString();
        saveTextBox->Select(saveTextBox->Text->Length(), 0);
    });
}

void DirectXPage::OnSaveAsTextBoxTextChanged(Platform::Object^ sender, Windows::UI::Xaml::Controls::TextChangedEventArgs^ e)
{
    saveButton->IsEnabled = safe_cast<TextBox^>(sender)->Text->Length() > 0;
}

void DirectXPage::OnSaveAsSaveButtonClick(Platform::Object^ sender, Windows::UI::Xaml::RoutedEventArgs^ e)
{
    saveAsFlyout->Hide();

    task<IVectorView<String^>^> getFilesTask = create_task(GetFileNamesAsync());
    getFilesTask.then([this](IVectorView<String^>^ fileNames)
    {
        String^ filename = saveTextBox->Text;
        unsigned int index;

        if (fileNames->IndexOf(filename, &index))
        {
            MessageDialog^ msgdlg = ref new MessageDialog("Replace existing drawing named '" + filename + "'?");
            msgdlg->Commands->Append(ref new UICommand("No", nullptr, false));
            msgdlg->Commands->Append(ref new UICommand("Yes", nullptr, true));
            msgdlg->CancelCommandIndex = 0;
            msgdlg->DefaultCommandIndex = 1;

            task<IUICommand^> messageTask = create_task(msgdlg->ShowAsync());
            messageTask.then([this, filename](IUICommand^ command)
            {
                if ((bool) command->Id)
                {
                    critical_section::scoped_lock lock(m_criticalSection);
                    IVector<StrokeInformation^>^ strokes = m_main->GetStrokes();

                    task<void> saveDrawingTask = create_task(SaveDrawingAsync(strokes, filename + DirectXPage::filenameExtension));
                    saveDrawingTask.then([this]()
                    {
                        appBarOpenButton->IsEnabled = true;
                    });
                };
            });
        }
        else
        {
            critical_section::scoped_lock lock(m_criticalSection);
            IVector<StrokeInformation^>^ strokes = m_main->GetStrokes();

            task<void> saveDrawingTask = create_task(SaveDrawingAsync(strokes, filename + DirectXPage::filenameExtension));
            saveDrawingTask.then([this]()
            {
                appBarOpenButton->IsEnabled = true;
            });
        }
    });
}


task<void> DirectXPage::SaveDrawingAsync(IVector<StrokeInformation^>^ strokes, String^ filename)
{
    StorageFolder^ folder = ApplicationData::Current->LocalFolder;

    task<StorageFile^> createFileTask = create_task(folder->CreateFileAsync(filename, CreationCollisionOption::ReplaceExisting));
    return createFileTask.then([](StorageFile^ file)
    {
        return file->OpenAsync(FileAccessMode::ReadWrite);
    })
        .then([this, strokes](IRandomAccessStream^ stream)
    {
        DataWriter^ writer = ref new DataWriter(stream);
        writer->WriteUInt32(strokes->Size);

        for (unsigned int i = 0; i < strokes->Size; i++)
        {
            StrokeInformation^ strokeInformation = strokes->GetAt(i);
            writer->WriteSingle(strokeInformation->Width);
            writer->WriteByte(strokeInformation->Color.A);
            writer->WriteByte(strokeInformation->Color.R);
            writer->WriteByte(strokeInformation->Color.G);
            writer->WriteByte(strokeInformation->Color.B);

            IVector<Point>^ points = strokeInformation->Points;
            writer->WriteUInt32(points->Size);

            for (unsigned int i = 0; i < points->Size; i++)
            {
                Point point = points->GetAt(i);
                writer->WriteSingle(point.X);
                writer->WriteSingle(point.Y);
            }
        }

        return writer->StoreAsync();
    })
        .then([](unsigned int)
    {
        return;
    });
}

task<IVectorView<StorageFile^>^> DirectXPage::GetFilesAsync()
{
    StorageFolder^ folder = ApplicationData::Current->LocalFolder;
    task<IVectorView<StorageFile^>^> getFilesTask = create_task(folder->GetFilesAsync(CommonFileQuery::OrderByName));
    return getFilesTask.then([](IVectorView<StorageFile^>^ storageFiles)
    {
        return storageFiles;
    });
}

task<IVectorView<String^>^> DirectXPage::GetFileNamesAsync()
{
    task<IVectorView<StorageFile^>^> getFilesTask = create_task(GetFilesAsync());
    return getFilesTask.then([](IVectorView<StorageFile^>^ storageFiles)
    {
        Vector<String^>^ names = ref new Vector<String^>();

        for (StorageFile^ storageFile : storageFiles)
            if (storageFile->FileType == DirectXPage::filenameExtension)
                names->Append(storageFile->DisplayName);

        return names->GetView();
    });
}

